!+
! Module: omfastshift_mod
!
! Authors:  Jonathan Rainger - MSSL
!           Vladimir Yershov - MSSL (vny@mssl.ucl.ac.uk)
!
! Description: shifts fast mode event list by tracking history shift
!
! Comments: For BPE binning think about rebinning image to centroided pixels and remove drift
!           rather than dividing drift by two.
!
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!-
! History:
!          1998 April: created jfr
!          1998 May 6: removed datestamp call (JFR)
!                      included readHist call (JFR)
!          1999 Jan 6: updated (PSW) - spr 555
!          2001 May 3 V.Ye. instrumentMode argument has been added to the getStartEndExp routine
!          2001 May 10 V.Ye. actual initial time of the observation has been added (obtStart)
!                      when calculating  On board Time for each event
!          2001 May 14 V.Ye. windowX0P and windowY0P have been added for calculation of zero-point of the scientific window
!          2001 May 16 V.Ye. TIMEZERO corrected to be the obs. start time relative to MJDREF (necessary for lcplot)
!          2001 July 11 V.Ye. Checking negative values in the output table (for the corrX and corrY values)
!          2001 September 28 V.Ye. - a bug in calculation of obtEvent has been fixed: obtEvent was already in seconds
!                                    elapsed since the mission reference time, hence a simple assignment is used
!                                    instead of call to OAL_obtToTimeTag(obtEvent, timeP);
!                                    ftFine calculation was also corrected.
!          2001 December 18, VNYe -  MJDREF keyword in the output temporary event list is set to 50814,
!                                    which corresponds to 1998.01.01 reference date (previosly it was erroneously
!                                    set to 50813) - SPR-2676.  
!          2002  January 07, SRR, VNY - TIMEZERO has been set to 0.d0 as it is relative to the parameter TSTART
!--------------------------------------------------------------------------------------------------

module omfastshift_mod

   use cal
   use oal
   use dal
   use param
   use errorHandling
   use getstartendexp_mod

   implicit none

   private

   public omfastshift

contains
  
!---------------------------------------------------------------------
subroutine omfastshift
!---------------------------------------------------------------------
    
   character(len=255) :: set                                         ! OM ODF Event List filename.
   character(len=255) :: thxSet                                      ! OM ODF Tracking History Data Auxiliary filename.
   character(len=255) :: nphSet                                      ! OM ODF Non-Periodic Housekeeping filename.
   logical(kind=bool) :: interpFlag                                  ! Interpolate tracking frame drift over fast timeslice?

   type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
   type(DataSetT) :: thxDataSetH                                     ! Tracking History Data dataset handle.

   type(TableT) :: inTableH                                          ! Input table handle.
   type(TableT) :: modesTableH                                       ! Modes table dataset handle.
   type(TableT) :: thxTableH                                         ! Tracking History Data table dataset handle.

   type(ColumnT), dimension(4) :: inColumnH                          ! Column handles.
   type(ColumnT), dimension(4) :: modesColumnH                       ! Modes column handle.
   type(ColumnT), dimension(2) :: thxColumnH                         ! Tracking History Data column handles.
   type(ColumnT), dimension(4) :: outColumnH                         ! Column handles.

   integer(kind=int8), dimension(:), pointer :: I8ColumnP            ! Integer 8 column pointer.
   integer(kind=int16), dimension(:), pointer :: I16ColumnP          ! Integer 16 column pointer.
   real(kind=Double), dimension(:), pointer :: R64ColumnP            ! Column pointer.

   character(len=255) :: dataMode                                    ! Input file data mode.

   integer(kind=int32) :: i                                          ! Counter.
   integer(kind=int32) :: nTrackFrames                               ! Number of tracking frames.
   integer(kind=int32) :: nEvents                                    ! Number of events.

   integer(kind=int32) :: frmTime                                    ! Frame time for tracking frames (in 1/1024 seconds)

   logical(kind=bool) :: binBPE                                         ! BPE binning?
   logical(kind=bool) :: zeroDrift                                      ! is there tracking history file? 

   integer(kind=int32), dimension(:), pointer :: ftCoarseP           ! Coarse time column pointer.
   integer(kind=int16), dimension(:), pointer :: ftFineP             ! Fine time column pointer.
   integer(kind=int16), dimension(:), pointer :: rawXP               ! Raw x column pointer.
   integer(kind=int16), dimension(:), pointer :: rawYP               ! Raw y column pointer.

   integer(kind=int32), dimension(:), pointer :: windowDXP           ! WindowDX pointer.
   integer(kind=int32), dimension(:), pointer :: windowDYP           ! WindowDY pointer.

   integer(kind=int32), dimension(:), pointer :: windowX0P           ! WindowX0 pointer.
   integer(kind=int32), dimension(:), pointer :: windowY0P           ! WindowY0 pointer.

   integer(kind=int32), dimension(:), pointer :: dxP                 ! Drift in X column pointer.
   integer(kind=int32), dimension(:), pointer :: dyP                 ! Drift in Y column pointer.

   real(kind=single), dimension(:), allocatable :: driftX            ! Drift in X.
   real(kind=single), dimension(:), allocatable :: driftY            ! Drift in Y.

   real(kind=single), dimension(:), allocatable :: incDriftX         ! Incremental drift in X.
   real(kind=single), dimension(:), allocatable :: incDriftY         ! Incremental drift in Y.
   real(kind=double) :: driftInterpX                                 ! Interpolated incremental drift in X.
   real(kind=double) :: driftInterpY                                 ! Interpolated incremental drift in Y.
   real(kind=double), dimension(:), allocatable :: obtTrackFrame     ! On board Time for tracking frames.

   real(kind=double) :: obtStart                                     ! On board time start.
   real(kind=double) :: obtEnd                                       ! On board time end.
   real(kind=double), dimension(:), allocatable :: obtEvent          ! On board time for event.

   integer(kind=int32) :: in                                         ! Frame Counter.
! VNY 28.09.2001 timeP changed to a simple allocatable array
!   real(kind=double), dimension(:), pointer :: timeP                 ! Time pointer.
   real(kind=double), dimension(:), allocatable :: timeP          ! time for event.

   integer(kind=int16), dimension(:), allocatable :: corrX           ! Drift corrected X pixel values.
   integer(kind=int16), dimension(:), allocatable :: corrY           ! Drift corrected Y pixel values.  
   integer(kind=int8), dimension(:), allocatable :: flag             ! Flag shifted event if outside science window.
   integer(kind=int32), dimension(1) :: columnDim                    ! Column dimension.

!******************************************************
! The following 2 variables have been added by 
! CB on 25.7.2000 for the call to getstartendexp
!*****************************************************
! The variable instrumentMode has been added (V.Ye. 03.05.2001) to getStartEndExp
!----------------------------------
   character(len=255) dateObs, dateEnd, obsID, expID
   character(len=3)   :: exposureNumber       ! String to hold exposure number
   character(len=3)   :: filterString
   integer            :: expNumber            ! Exposure number
   integer            :: instrumentMode = DATA_FAST       ! string for the instrument mode (IMAGE, DATA_FAST, etc.)
 
   set = stringParameter("set")
   thxSet = stringParameter("thxset")
   nphSet = stringParameter("nphset")
   interpFlag = booleanParameter("interpolate")

! Get a hold on the ODF OM Tracking History Data Auxiliary File.

   thxDataSetH = dataSet(thxSet, read)

! Determine file type from file name in primary header.

!---Added VNY 04.07.2001 in order to treat the absense of THX-file
   zeroDrift = .false.
   if (hasBlock(thxDataSetH, "ZERODRIFT")) then
      zeroDrift = .true.
      dataMode="FAST"
   else


      dataMode = trim(stringAttribute(thxDataSetH, "DATAMODE")) 
   
      if (dataMode == "TRACKING") then
     
         ! Get a handle on the tracking history data binary table extension.
      
         thxTableH = table(thxDataSetH, "OMTHX1")
      
         nTrackFrames = numberOfRows(thxTableH)
      
         thxColumnH(1) = column(thxTableH, "DX", read)
         dXP => int32Data(thxColumnH(1))
      
         thxColumnH(2) = column(thxTableH, "DY", read)
         dYP => int32Data(thxColumnH(2))
      
         ! Allocate drift arrays.

         allocate(driftX(1:nTrackFrames))
         allocate(driftY(1:nTrackFrames))
         allocate(incDriftX(1:nTrackFrames))
         allocate(incDriftY(1:nTrackFrames))
         allocate(obtTrackFrame(1:nTrackFrames))

         ! Store integer drifts as real drifts

         driftX = real(dXP)/1000.0
         driftY = real(dYP)/1000.0

         ! Calculate drifts.

         incDriftX(1) = real(dXP(0))/1000.0
         incDriftY(1) = real(dYP(0))/1000.0
      
         do i=2,nTrackFrames

            incDriftX(i) = real(dXP(i-1) - dXP(i-2))/1000.0
            incDriftY(i) = real(dYP(i-1) - dYP(i-2))/1000.0

         end do



         ! If BPE binning implemented ...

         if (binBPE) then
            driftX = driftX/2.0
            driftY = driftY/2.0
            incDriftX = incDriftX/2.0
            incDriftY = incDriftY/2.0
         end if
         
         ! Release handles on column and table.

         call release(thxColumnH(1))
         call release(thxColumnH(2))
         call release(thxTableH)
         call release(thxDataSetH)
      else 

         ! Incompatible file.

         call fatal("badFileMode","Value for keyword DATAMODE is incompatible")

      end if !dataMode=Tracking
   end if ! hasBlock ZERODRIFT
   
! Read in ODF Event List file.

   inDataSetH = dataSet(set, modify)


! Determine file type from file name in primary header.

   dataMode = trim(stringAttribute(inDataSetH, "DATAMODE"))
   binBPE = booleanAttribute(inDataSetH, "BINBPE")
   frmTime =  int32Attribute(inDataSetH, "FRMTIME")

   if (dataMode == "FAST") then

! OM Fast Mode Event List File. 

! Get a handle on the event list binary table extension.

      inTableH = table(inDataSetH, "OMFAE1")
      
      inColumnH(1) = column(inTableH, "FTCOARSE", read)
      ftCoarseP => int32Data(inColumnH(1))
      
      inColumnH(2) = column(inTableH, "FTFINE", read)
      ftFineP => int16Data(inColumnH(2))
      
      inColumnH(3) = column(inTableH, "RAWX", read)
      rawXP => int16Data(inColumnH(3))
      
      inColumnH(4) = column(inTableH, "RAWY", read)
      rawYP => int16Data(inColumnH(4))
      
! Get the number of rows in the event list binary table extension.

      nEvents = numberOfRows(inTableH)

! Read modes extension.

      modesTableH = table(inDataSetH, "MODES")

!------ Added by V.Ye. 14.05.2001 windowX0 and windowY0 (zero-point of the scientific window)

      modesColumnH(1) = column(modesTableH, "WINDOWX0", read)
      windowX0P => int32data(modesColumnH(1))      

      modesColumnH(2) = column(modesTableH, "WINDOWY0", read)
      windowY0P => int32data(modesColumnH(2))
!------------
      
      modesColumnH(3) = column(modesTableH, "WINDOWDX", read)
      windowDXP => int32data(modesColumnH(3))
      
      modesColumnH(4) = column(modesTableH, "WINDOWDY", read)
      windowDYP => int32data(modesColumnH(4))

!***********************************************************
! The following 2 lines have been added by CB on
! 25.7.2000
!***********************************************************
      dateObs = stringAttribute(inTableH, "DATE-OBS");
      dateEnd = stringAttribute(inTableH, "DATE-END"); 
      obsID = stringAttribute(inTableH, "OBS_ID");
      expID = stringAttribute(inTableH, "EXP_ID"); 

   else 
      call fatal("unkownMode","Unknown data mode")
   end if  ! dataMode=FAST

! allocate arrays

!--- VNY 11.07.2001 the following allocation should be done AFTER cheking  the events are not negative 
   allocate(corrX(1:nEvents))
   allocate(corrY(1:nEvents))
   allocate(flag(1:nEvents))
   allocate(obtEvent(nEvents))
!---VNY 28.09.2001 timeP is allocated as well
   allocate(timeP(nEvents))

! Get start and end time of exposure from the non-periodic housekeeping file.

!********************************************************
! The new call to getStartEndExp - CB 25.7.2000
! The following is taken from omprep_mod.f90 (AMK code)
!********************************************************
! The new call to getStartEndExp - V.Ye. 3.05.2001
! the instrumentMode argument has been added
!------------------------------------------------
    exposureNumber = expId(len_trim(obsId) + 1 : len_trim(expId)) 

    ! Now convert the string to an integer
    read(exposureNumber, '(i3)') expNumber

    ! Now find the accurate start and end times for this exposure
    ! NB filter string has been added on 4.10.2000 -= but noit initialsed
    call getStartEndExp(nphSet, expNumber, instrumentMode, obtStart, obtEnd, dateObs, dateEnd, filterString)

! Calculate tracking frame time stamps in OBT for middle of tracking frame.
    if (.not. zeroDrift) then 
       do i = 1, nTrackFrames
          obtTrackFrame(i) = obtStart + (REAL(i) - 0.5) * REAL(frmTime) / 1024.0
       end do
    endif


! Do 'shift and add' process.

   do i = 1, nEvents

! Calculate On board Time for each event taking into acount wrap around for int16 values.


!---VNY 25.09.2001 Test of a different ftFine calculation 
      obtEvent(i) =  REAL(ftCoarseP(i-1) - ftCoarseP(0)) + obtStart + REAL(ftFineP(i-1)) / 1024.     

! Locate tracking frame in which event occured.

      if (.not. zeroDrift) then
         call locate(obtTrackFrame, nTrackFrames, obtEvent(i), in)

         ! If linear interpolation of drifts is required ...


         if (interpFlag) then
 
            if (in == 1) then

               driftInterpX = 0.0 + ((obtEvent(i) - obtStart)/((frmTime/1024.0)/2.0))*incDriftX(in) 
               driftInterpY = 0.0 + ((obtEvent(i) - obtStart)/((frmTime/1024.0)/2.0))*incDriftY(in)

            else

               driftInterpX = driftX(in-1) + ((obtEvent(i) - obtTrackFrame(in-1))/(frmTime/1024.0))*incDriftX(in) 
               driftInterpY = driftY(in-1) + ((obtEvent(i) - obtTrackFrame(in-1))/(frmTime/1024.0))*incDriftY(in)

            end if


            corrX(i) = floor(real(rawXP(i-1)) + 0.5 - driftInterpX)
            corrY(i) = floor(real(rawYP(i-1)) + 0.5 - driftInterpY)

            ! no linear interpolation of drift required ...

         else
  
            corrX(i) = floor(real(rawXP(i-1)) + 0.5 - driftX(in))
            corrY(i) = floor(real(rawYP(i-1)) + 0.5 - driftY(in))

         end if
      else
!--- Added VNY 04.07.2001 in order to deal with the absense of tracking history file
         corrX(i) = real(rawXP(i-1))
         corrY(i) = real(rawYP(i-1))
      end if

! Is drifted event in Science window??

      if (corrX(i) < windowX0P(0) .or. corrX(i) >= windowX0P(0)+windowDXP(0) .or. &
           corrY(i) < windowY0P(0) .or. corrY(i) >= windowY0P(0)+windowDYP(0)) then
         flag(i) = 1
      else
         flag(i) = 0
      end if
   end do


! Write time attributes to input file.

!--- SRR (19/12/01) TIMEZERO has been set to 0.d0 as it is relative to the parameter TSTART

   call setAttribute(inDataSetH, "TIMESYS", "TT", "fundamental time system: Terrestial Time")
   call setAttribute(inDataSetH, "TIMEUNIT", "s", "unit for TSTART, TSTOP, TIMEZERO, TELPASE")
   call setAttribute(inDataSetH, "TASSIGN", "SATELLITE", "location of time assignment")
   call setAttribute(inDataSetH, "TIMEREF", "LOCAL", "spatial reference frame")
   call setAttribute(inDataSetH, "CLOCKAPP", "T","times corrected for clock drifts?")
!   call setAttribute(inDataSetH, "TIMEZERO", obtStart, "s", "obs. start time relative to MJDREF")
   call setAttribute(inDataSetH, "TIMEZERO", 0.d0, "s", "obs. start time relative to MJDREF")
   call setAttribute(inDataSetH, "MJDREF", 50814.000731296D0, "d", "1998.001 (UTC) in TT")

! Set the dimensions of the output columns.

   columnDim(1) = 1

! Add new columns.

! Convert Onboard Timestamps (OBT) to event time tags (elapsed seconds since mission 
! reference time 1998-01-01T00:00:00).
   
!---VNY 28.09.2001 obtEvent is already in seconds elapsed since the mission reference time
!   hence a simple assignment is used instead of call to OAL_obtToTimeTag(obtEvent, timeP)
!   call OAL_obtToTimeTag(obtEvent, timeP)
   timeP=obtEvent


   outColumnH(1) = addColumn(inTableH, "TIME", real64, "seconds", columnDim)
   R64ColumnP => real64data(outColumnH(1))
   R64ColumnP = timeP

   outColumnH(2) = addColumn(inTableH, "CORR_X", integer16, "pixel", columnDim)
   I16ColumnP => int16data(outColumnH(2))
   I16ColumnP = corrX

   call  setAttribute(outColumnH(2), "TLMIN", 0)
   call setAttribute(outColumnH(2), "TLMAX", windowDXP(0)-1)

   outColumnH(3) = addColumn(inTableH, "CORR_Y", integer16, "pixel", columnDim)
   I16ColumnP => int16data(outColumnH(3))
   I16ColumnP = corrY

   call setAttribute(outColumnH(3), "TLMIN", 0)
   call setAttribute(outColumnH(3), "TLMAX", windowDYP(0)-1)

   outColumnH(4) = addColumn(inTableH, "WIN_FLAG", integer8, "", columnDim)
   I8ColumnP => int8data(outColumnH(4))
   I8ColumnP = flag

   call release(inColumnH(1))
   call release(inColumnH(2))
   call release(inColumnH(3))
   call release(inColumnH(4))
   call release(modesColumnH(1))
   call release(modesColumnH(2))
   call release(modestableH)
   call release(outColumnH(1))
   call release(outColumnH(2))
   call release(outColumnH(3))
   call release(outColumnH(4))
   call release(inTableH)
   call release(inDataSetH)

end subroutine omfastshift

!---------------------------------------------------------------------
subroutine locate(xx, n, x, j)
!---------------------------------------------------------------------

   use dal

   integer(kind=int32), intent(in) :: n
   real(kind=double), intent(in), dimension(n) :: xx
   real(kind=double), intent(in) :: x
   integer(kind=int32), intent(out) :: j

   integer(kind=int32) ju
   integer(kind=int32) jl
   integer(kind=int32) jm

   j = 1

   if (x < xx(1)) then
      j = 0
   end if

   if (x > xx(n)) then
      j = n - 1
   end if

   if ( j == 1) then
      jl = 0
      ju = n + 1

      do while (ju - jl > 1)
         jm = (ju + jl) / 2
         if ((xx(n) > xx(1)) .eqv. (x > xx(jm))) then
            jl = jm
         else
            ju = jm
         end if
      end do

      j = jl
   end if
  
   j = j + 1  

end subroutine locate

end module omfastshift_mod


